/***************************************************************************
 *   Copyright (C) 2008  Antony Martin <martin(dot)antony(at)yahoo(dot)fr> *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU General Public License           *
 *   as published by the Free Software Foundation; either version 2        *
 *   of the License, or (at your option) any later version.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA          *
 ***************************************************************************/

/* Cree le : 19 janvier 2007
   derniere modification le 12/07/2008 */

#include <SCE/SCEMinimal.h>
#include <SCE/interface/lib4fm.h>
#include <libwar/libwar.h>

#include <SCE/utils/SCEVector.h>
#include <SCE/utils/SCEString.h>
#include <SCE/utils/SCEResources.h>
#include <SCE/utils/SCEMedia.h>

#include <SCE/core/SCECTexture.h>
#include <SCE/core/SCECBuffers.h>

#include <SCE/interface/SCEMesh.h>

#include <SCE/interface/SCE4FMLoader.h>
#include <SCE/interface/SCEOBJLoader.h>


static int is_init = SCE_FALSE;

/* mesh en cache pour le dessin */
static SCE_SMesh *m = NULL;

static int sce_4fm_media_type;
static int sce_obj_media_type;

/* liste des mises a jour a effectuer */
static SCE_SList *updates = NULL;
/* pointeur sur le dernier tampon (SCE_SMeshVertexBuffer)
   dont on a demande la mise a jour */
static SCE_SListIterator *upit = NULL;
/* nombre de mises a jour demandees */
static unsigned int num_updates = 0;

/* revise le 26/03/2008 */
int SCE_Init_Mesh (void)
{
    SCE_btstart ();
    if (!is_init)
    {
        if (!(updates = SCE_List_Create (NULL)))
        {
            Logger_LogSrc ();
            return SCE_ERROR;
        }

        /* enregistrement des loaders par defaut */
        sce_4fm_media_type = SCE_Media_GenTypeID ();
        SCE_Media_RegisterLoader (sce_4fm_media_type, FFM_MAGIC,
                                  "."FFM_FILE_EXTENSION, SCE_4FM_Load);
        sce_obj_media_type = SCE_Media_GenTypeID ();
        SCE_Media_RegisterLoader (sce_obj_media_type, 0,
                                  "."WAR_FILE_EXTENSION, SCE_OBJ_Load);

        is_init = SCE_TRUE;
    }
    SCE_btend ();
    return SCE_OK;
}

/* revise le 26/03/2008 */
void SCE_Quit_Mesh (void)
{
    is_init = SCE_FALSE;
    SCE_List_Delete (updates), updates = NULL;
    num_updates = 0;
    upit = NULL;
}


/* ajoute le 14/10/2007 */
static void SCE_Mesh_InitVertexData (SCE_SMeshVertexData *d)
{
    d->dec = NULL;
    d->data = NULL;
    d->canfree = SCE_FALSE;
    d->size = 0;
}
/* ajoute le 14/10/2007 */
static SCE_SMeshVertexData* SCE_Mesh_CreateVertexData (void)
{
    SCE_SMeshVertexData *d = NULL;

    SCE_btstart ();
    d = SCE_malloc (sizeof *d);
    if (!d)
    {
        Logger_LogSrc ();
        SCE_btend ();
        return NULL;
    }

    SCE_Mesh_InitVertexData (d);
    d->dec = SCE_CCreateVertexDeclaration ();
    if (!d->dec)
    {
        SCE_free (d);
        Logger_LogSrc ();
        SCE_btend ();
        return NULL;
    }

    SCE_btend ();
    return d;
}
/* ajoute le 14/10/2007 */
/* revise le 06/01/2008 */
static void SCE_Mesh_DeleteVertexData (void *data)
{
    SCE_btstart ();
    if (data)
    {
        SCE_SMeshVertexData *d = data;
        SCE_CDeleteVertexDeclaration (d->dec);
        if (d->canfree)
            SCE_free (d->data);
        SCE_free (d);
    }
    SCE_btend ();
}

/* revise le 13/10/2007 */
void SCE_Mesh_InitVB (SCE_SMeshVertexBuffer *vb)
{
    vb->usage = GL_STREAM_DRAW;
    vb->buffer = NULL;
    vb->data = NULL;
    vb->active = SCE_FALSE;
    vb->builded = SCE_FALSE;
}

/* ajoute le 01/01/2008 */
void SCE_Mesh_InitIB (SCE_SMeshIndexBuffer *ib)
{
    ib->usage = GL_STATIC_DRAW;
    ib->buffer = NULL;
    ib->type = SCE_UNSIGNED_BYTE;
    ib->data = NULL;
    ib->canfree = SCE_FALSE;
    ib->size = 0;
    ib->builded = SCE_FALSE;
}

/* ajoute le 13/10/2007 */
/* revise le 26/03/2008 */
SCE_SMeshVertexBuffer* SCE_Mesh_CreateVB (void)
{
    SCE_SMeshVertexBuffer *vb = NULL;

    SCE_btstart ();
    vb = SCE_malloc (sizeof *vb);
    if (!vb)
        Logger_LogSrc ();
    else
    {
#define SCE_MESHASSERT(c)\
        if (c)\
        {\
            SCE_Mesh_DeleteVB (vb);\
            Logger_LogSrc ();\
            SCE_btend ();\
            return NULL;\
        }
        /* pour chaque buffer cree, on lui reserve une place
           dans la liste des buffers a mettre a jour */
        SCE_MESHASSERT (SCE_List_AddBack (updates, NULL) < 0)
        SCE_Mesh_InitVB (vb);
        SCE_MESHASSERT (!(vb->buffer = SCE_CCreateVertexBuffer ()))
        vb->data = SCE_List_Create (SCE_Mesh_DeleteVertexData);
        if (!vb->data)
        {
            SCE_Mesh_DeleteVB (vb), vb = NULL;
            Logger_LogSrc ();
        }
#undef SCE_MESHASSERT
    }
    SCE_btend ();
    return vb;
}
/* ajoute le 13/10/2007 */
/* revise le 17/10/2007 */
void SCE_Mesh_DeleteVB (void *b)
{
    SCE_btstart ();
    if (b)
    {
        SCE_SMeshVertexBuffer *vb = b;
        SCE_CDeleteVertexBuffer (vb->buffer);
        SCE_List_Delete (vb->data);
        SCE_free (vb);
    }
    SCE_btend ();
}

/* ajoute le 16/10/2007 */
/* revise le 06/01/2008 */
int SCE_Mesh_BuildVB (SCE_SMeshVertexBuffer *vb)
{
    SCE_SListIterator *i = NULL;
    SCE_SMeshVertexData *d = NULL;

    SCE_btstart ();
    /* si le buffer est deja construit ou si il ne possede aucune donnee */
    if (vb->builded || SCE_List_GetSize (vb->data) < 1)
    {
        SCE_btend ();
        return SCE_OK;
    }

    SCE_CBindVertexBuffer (vb->buffer);
    SCE_List_ForEach (i, vb->data)
    {
        d = SCE_List_GetData (i);
        d->dec->first = SCE_CGetVertexBufferFirst (vb->buffer);
        if (SCE_CAddVertexDeclaration_ (d->dec) < 0)
        {
            SCE_CBindVertexBuffer (NULL);
            Logger_LogSrc ();
            SCE_btend ();
            return SCE_ERROR;
        }
        if (SCE_CAddVertexBufferData_ (d->size, d->data) < 0)
        {
            SCE_CBindVertexBuffer (NULL);
            Logger_LogSrc ();
            SCE_btend ();
            return SCE_ERROR;
        }
    }
    SCE_CBuildVertexBuffer_ (vb->usage);
    SCE_CBindVertexBuffer (NULL);

    SCE_btend ();
    return SCE_OK;
}

/* ajoute le 16/10/2007 */
/* revise le 24/10/2007 */
int SCE_Mesh_BuildIB (SCE_SMeshIndexBuffer *ib)
{
    SCE_btstart ();
    if (!ib->builded && ib->data)
    {
        SCE_CBindIndexBuffer (ib->buffer);
        if (SCE_CAddIndexBufferData_ (ib->size, ib->data) < 0)
        {
            SCE_CBindIndexBuffer (NULL);
            Logger_LogSrc ();
            SCE_btend ();
            return SCE_ERROR;
        }
        SCE_CBuildIndexBuffer_ (ib->usage);
        SCE_CBindIndexBuffer (NULL);
    }
    SCE_btend ();
    return SCE_OK;
}


/* revise le 01/01/2008 */
void SCE_Mesh_Init (SCE_SMesh *mesh)
{
    mesh->vertices = NULL;

    SCE_Mesh_InitIB (&mesh->ib);

    mesh->vcount = 0;
    mesh->icount = 0;
    mesh->use_indices = SCE_FALSE;
    mesh->polygon_type = GL_POINTS; /* evitons tout probleme */
    mesh->builded = SCE_FALSE;

    mesh->pos = mesh->nor = NULL;

    mesh->sortedfaces = NULL;
}

/* revise le 13/10/2007 */
SCE_SMesh* SCE_Mesh_Create (void)
{
    SCE_SMesh *mesh = NULL;

    SCE_btstart ();
    mesh = SCE_malloc (sizeof *mesh);
    if (!mesh)
    {
        Logger_LogSrc ();
        SCE_btend ();
        return NULL;
    }

    SCE_Mesh_Init (mesh);
    mesh->ib.buffer = SCE_CCreateIndexBuffer ();
    if (!mesh->ib.buffer)
    {
        SCE_free (mesh);
        Logger_LogSrc ();
        SCE_btend ();
        return NULL;
    }
    mesh->vertices = SCE_List_Create (SCE_Mesh_DeleteVB);
    if (!mesh->vertices)
    {
        SCE_Mesh_Delete (mesh), mesh = NULL;
        Logger_LogSrc ();
    }

    SCE_btend ();
    return mesh;
}

/* revise le 14/02/2008 */
void SCE_Mesh_Delete (SCE_SMesh *mesh)
{
    SCE_btstart ();
    if (mesh && SCE_Resource_Free (mesh))
    {
        if (mesh->ib.canfree)
            SCE_free (mesh->ib.data);
        SCE_CDeleteIndexBuffer (mesh->ib.buffer);
        SCE_List_Delete (mesh->vertices);
        SCE_List_Delete (mesh->sortedfaces);
        SCE_free (mesh);
    }
    SCE_btend ();
}
/* ajoute le 17/02/2008 */
void SCE_Mesh_DeleteList (SCE_SMesh **meshs)
{
    if (meshs && SCE_Resource_Free (meshs))
    {
        unsigned int i = 0;
        while (meshs[i])
            SCE_Mesh_Delete (meshs[i++]);
        SCE_free (meshs);
    }
}


/* ajoute le 21/10/2007 */
int SCE_Mesh_AddVB (SCE_SMesh *mesh, SCE_SMeshVertexBuffer *b)
{
    SCE_btstart ();
    if (SCE_List_AddBack (mesh->vertices, b) < 0)
    {
        Logger_LogSrc ();
        SCE_btend ();
        return SCE_ERROR;
    }
    SCE_btend ();
    return SCE_List_GetSize (mesh->vertices) - 1;
}
/* ajoute le 14/10/2007 */
int SCE_Mesh_AddNewVB (SCE_SMesh *mesh, int mode)
{
    int ret = SCE_ERROR;
    SCE_SMeshVertexBuffer *b = NULL;

    SCE_btstart ();
    b = SCE_Mesh_CreateVB ();
    if (!b)
    {
        Logger_LogSrc ();
        SCE_btend ();
        return SCE_ERROR;
    }
    /* mode peut etre egal a zero pour laisser la valeur par defaut */
    if (mode)
        b->usage = mode;
    ret = SCE_Mesh_AddVB (mesh, b);
    if (ret < 0)
    {
        SCE_Mesh_DeleteVB (b);
        Logger_LogSrc ();
    }
    SCE_btend ();
    return ret;
}
/* ajoute le 14/10/2007 */
SCE_SMeshVertexBuffer* SCE_Mesh_RemoveVB (SCE_SMesh *mesh, unsigned int id)
{
    SCE_SListIterator *i = NULL;

    SCE_btstart ();
    i = SCE_List_GetIterator (mesh->vertices, id);
    if (i)
    {
        SCE_btend ();
        return SCE_List_Remove (mesh->vertices, i);
    }
    else
    {
        Logger_LogSrc ();
        SCE_btend ();
        return NULL;
    }
}
/* ajoute le 14/10/2007 */
void SCE_Mesh_EraseVB (SCE_SMesh *mesh, unsigned int id)
{
    SCE_btstart ();
    SCE_Mesh_DeleteVB (SCE_Mesh_RemoveVB (mesh, id));
    SCE_btend ();
}


/* revise le 25/10/2007 */
int SCE_Mesh_ActiveVB (SCE_SMesh *mesh, unsigned int id, int active)
{
    SCE_SMeshVertexBuffer *vb = NULL;
    SCE_SListIterator *i = NULL;

    SCE_btstart ();
    if (!mesh)
        mesh = m;

    i = SCE_List_GetIterator (mesh->vertices, id);
    if (!i)
        Logger_Clear (); /* c'est juste un mauvais 'id' ... */
    else
    {
        vb = SCE_List_GetData (i);
        vb->active = active;
    }
    SCE_btend ();
    return SCE_OK;
}

/* revise le 13/10/2007 */
void SCE_Mesh_ActiveIB (SCE_SMesh *mesh, int active)
{
    if (mesh)
        mesh->use_indices = active;
    else
        m->use_indices = active;
}


/* ajoute le 14/10/2007 */
/* revise le 16/02/2008 */
/* recherche et renvoie les donnees de 'mesh' ('vertices')
   ayant pour attribut 'attrib'. ecrit egalement l'adresse
   du buffer correspondant dans 'vb' */
SCE_SMeshVertexData* SCE_Mesh_LocateData (SCE_SMesh *mesh, int attrib,
                                          SCE_SMeshVertexBuffer **vb)
{
    SCE_SListIterator *i = NULL, *j = NULL;
    SCE_SMeshVertexData *data = NULL;
    SCE_SMeshVertexBuffer *tmp = NULL;

    SCE_btstart ();
    if (!vb)
        vb = &tmp;

    SCE_List_ForEach (i, mesh->vertices)
    {
        *vb = SCE_List_GetData (i);
        SCE_List_ForEach (j, (*vb)->data)
        {
            data = SCE_List_GetData (j);
            if (data->dec->attrib == attrib)
            {
                SCE_btend ();
                return data;
            }
        }
    }

    *vb = NULL;
    SCE_btend ();
    return NULL;
}

/* ajoute le 05/05/2008 */
void* SCE_Mesh_GetVerticesPositions (SCE_SMesh *mesh)
{
    return mesh->pos;
}
void* SCE_Mesh_GetVerticesNormals (SCE_SMesh *mesh)
{
    return mesh->nor;
}

/**
Attention : l'utilisateur peut envoyer n'importe quel type de donnees a
            AddVertices...
**/
/* revise le 06/01/2007 */
int SCE_Mesh_AddVertices (SCE_SMesh *mesh, unsigned int id, int attrib,
                          SCEenum type, unsigned int size, unsigned int count,
                          void *data, int canfree)
{
    SCE_SMeshVertexData *d = NULL;
    SCE_SListIterator *it = NULL;
    SCE_SMeshVertexBuffer *vb = NULL;

    SCE_btstart ();
    if (!data)
    {
        SCE_btend ();
        return SCE_OK;
    }

    /* verification du count (si un count est deja present) */
    if (!mesh->vcount)
        mesh->vcount = count;
    else if (count < mesh->vcount)
    {
        Logger_Log (SCE_INVALID_ARG);
        Logger_LogMsg ("too small count, the minimum requested for"
                       " this mesh is %u", mesh->vcount);
        SCE_btend ();
        return SCE_ERROR;
    }

    /* on verifie si ce type de donnee a deja ete specifie */
    if ((d = SCE_Mesh_LocateData (mesh, attrib, &vb)))
    {
        /* si ces donnees existent, on les supprime dans leur buffer */
        it = SCE_List_LocateIterator (vb->data, d, NULL);
        SCE_List_Remove (vb->data, it);
        it = NULL;
        if (d->canfree)
            SCE_free (d->data);
        d->data = NULL;
        /* et on demande la reconstruction du buffer */
        SCE_CClearVertexBuffer (vb->buffer);
        vb->builded = SCE_FALSE;
        /* si le buffer est vide et qu'il ne s'agit pas
           du buffer de destination des donnees */
        if (SCE_List_GetSize (vb->data) < 1 &&
            SCE_List_LocateIndex (mesh->vertices, vb, NULL) != id)
        {
            /* on peut le supprimer */
            it = SCE_List_LocateIterator (mesh->vertices, vb, NULL);
            SCE_List_Erase (mesh->vertices, it);
        }
    }
    else
    {
        /* creation d'une donnee de vertex buffer */
        d = SCE_Mesh_CreateVertexData ();
        if (!d)
        {
            Logger_LogSrc ();
	    SCE_btend ();
            return SCE_ERROR;
        }
    }

    /* recuperation du vb... */
    it = SCE_List_GetIterator (mesh->vertices, id);
    if (!it)
    {
        /* ...creation si inexistant */
        Logger_Clear (); /* suppression de l'erreur loggee par GetIterator */
        if ((id = SCE_Mesh_AddNewVB (mesh, 0)) < 0)
        {
            Logger_LogSrc ();
            SCE_btend ();
            return SCE_ERROR;
        }
        /* ce Get ne peut pas echouer */
        it = SCE_List_GetIterator (mesh->vertices, id);
    }
    vb = SCE_List_GetData (it);

    if (vb->builded)
    {
        /* la reconstruction du buffer sera requise */
        SCE_CClearVertexBuffer (vb->buffer);
        vb->builded = SCE_FALSE;
    }

    /* ajout des donnees au buffer */
    if (SCE_List_AddBack (vb->data, d) < 0)
    {
        SCE_Mesh_DeleteVertexData (d);
        Logger_LogSrc ();
        SCE_btend ();
        return SCE_ERROR;
    }

    /* specification des donnes */
    d->dec->attrib = attrib;
    d->dec->type = type;
    d->dec->size = size;

    d->size = count * size * SCE_CSizeofType (type);
    /* aucune duplication des donnees */
    d->data = data;
    d->canfree = canfree;

    if (attrib == SCE_POSITION)
        mesh->pos = data; /* hopla */
    else if (attrib == SCE_NORMAL)
        mesh->nor = data; /* hopli */

    mesh->builded = SCE_FALSE;

    SCE_btend ();
    /* par defaut on retourne l'id du tampon,
       au cas ou on en aurait cree un nouveau avec un id different */
    return id;
}

/* ajoute le 17/10/2007 */
/* revise le 06/01/2008 */
int SCE_Mesh_AddVerticesDup (SCE_SMesh *mesh, unsigned int id, int attrib,
                             SCEenum type, unsigned int size,
                             unsigned int count, void *data)
{
    /* conversion des donnees */
    SCEvertices *new = NULL;

    SCE_btstart ();
    new = SCE_Mem_ConvertDup (SCE_VERTICES_TYPE, type, data, count * size);
    if (!new)
    {
        Logger_LogSrc ();
        SCE_btend ();
        return SCE_ERROR;
    }
    SCE_btend ();
    return SCE_Mesh_AddVertices (mesh, id, attrib, SCE_VERTICES_TYPE,
                                 size, count, new, SCE_TRUE);
}


/* revise le 06/01/2008 */
int SCE_Mesh_SetIndices (SCE_SMesh *mesh, SCEenum usage, SCEenum type,
                         unsigned int count, void *data, int canfree)
{
    SCE_btstart ();

    if (mesh->ib.builded)
    {
        /* suppression du buffer, car une reconstruction sera necessaire */
        SCE_CClearIndexBuffer (mesh->ib.buffer);
    }

    mesh->ib.size = count * SCE_CSizeofType (type);
    if (mesh->ib.canfree)
        SCE_free (mesh->ib.data); /* suppression des donnees precedentes */
    mesh->ib.data = data;
    mesh->ib.canfree = canfree;
    mesh->ib.type = type;
    if (usage)
        mesh->ib.usage = usage;

    mesh->use_indices = SCE_TRUE;
    mesh->icount = count;

    mesh->ib.builded = mesh->builded = SCE_FALSE;

    SCE_btend ();
    return SCE_OK;
}

/* ajoute le 20/10/2007 */
/* revise le 12/07/2008 */
int SCE_Mesh_SetIndicesDup (SCE_SMesh *mesh, SCEenum mode, SCEenum type,
                            unsigned int count, void *data)
{
    SCEenum t;
    void *data_dup = NULL;

    SCE_btstart ();
#if 0
    /* conversion des donnees en le type le plus leger */
    t = ((count >= 256*256) ? SCE_UNSIGNED_INT : SCE_UNSIGNED_SHORT);
#endif
    t = SCE_INDICES_TYPE;
    if (!(data_dup = SCE_Mem_ConvertDup (t, type, data, count)))
    {
        Logger_LogSrc ();
        SCE_btend ();
        return SCE_ERROR;
    }
    SCE_btend ();
    return SCE_Mesh_SetIndices (mesh, mode, t, count, data_dup, SCE_TRUE);
}


/* ajoute le 27/03/2008 */
int SCE_Mesh_GetNumVerticesPerFace (SCE_SMesh *mesh)
{
    int vpf;
    switch (mesh->polygon_type)
    {
    case SCE_POINTS:         vpf = 1; break;
    case SCE_LINES:          vpf = 2; break;/*
    case SCE_TRIANGLE_STRIP:
    case SCE_TRIANGLE_FAN:*/
    case SCE_TRIANGLES:      vpf = 3; break;
    case SCE_QUADS:          vpf = 4;
#ifdef SCE_DEBUG
        break;
    default:
        Logger_Log (-1);
        Logger_LogMsg ("unsupported face type");
        return SCE_ERROR;
#endif
    }
    return vpf;
}
/* ajoute le 27/03/2008 */
int SCE_Mesh_GetNumFaces (SCE_SMesh *mesh)
{
    if (mesh->use_indices)
        return mesh->icount/SCE_Mesh_GetNumVerticesPerFace (mesh);
    else
        return mesh->vcount/SCE_Mesh_GetNumVerticesPerFace (mesh);
}

/* ajoute le 27/03/2008 */
int SCE_Mesh_ForEachFace (SCE_SMesh *mesh, SCE_FMeshFaceFunc func, void *param)
{
    SCE_SMeshFace face;
    int i, j, num_faces, vpf;

    num_faces = SCE_Mesh_GetNumFaces (mesh);
    vpf = SCE_Mesh_GetNumVerticesPerFace (mesh);
    face.num_vertices = vpf;
    face.pos = mesh->pos;
    face.nor = mesh->nor;

    if (mesh->use_indices)
    {
        unsigned int *indices = mesh->ib.data;
#ifdef SCE_DEBUG
        if (mesh->ib.type != SCE_UNSIGNED_INT)
        {
            Logger_Log (SCE_INVALID_OPERATION);
            Logger_LogMsg ("unsigned int indices expected on "
                           "this mesh for calling %s", __FUNCTION__);
            return SCE_ERROR;
        }
#endif
        for (i=0; i<num_faces; i++)
        {
            for (j=0; j<vpf; j++)
                face.index[j] = indices[i*vpf+j]*3;
            if (!func (mesh, &face, i, param))
                return SCE_OK;
        }
    }
    else
    {
        for (i=0; i<num_faces; i++)
        {
            for (j=0; j<vpf; j++)
                face.index[j] = (i*vpf+j)*3;
            if (!func (mesh, &face, i, param))
                return SCE_OK;
        }
    }
    return SCE_OK;
}


/* ajoute le 27/03/2008 */
/* calcul la position moyenne d'un polygone */
static void SCE_Mesh_GetFaceCenter (SCE_TVector3 v[4], int vpf, SCE_TVector3 c)
{
    int i;
    SCE_Vector3_Copy (c, v[0]);
    for (i=1; i<vpf; i++)
        SCE_Vector3_Operator1v (c, +=, v[i]);
    SCE_Vector3_Operator1 (c, /=, vpf);
}

/* structure d'une face vue par la liste 'sortedfaces' des meshs */
typedef struct
{
    float dist;
    SCEindices index[4];
} SCE_SMeshFaceInfo;
/* callback pour 'sortedfaces' pour supprimer une face (cf au ci-dessus) */
static void SCE_Mesh_DeleteFaceInfo (void *i)
{
    SCE_free (i);
}

/* ajoute une nouvelle face a la liste en copiant ses donnees */
static int SCE_Mesh_CopyFaceToSortedList (SCE_SMesh *mesh, SCE_SMeshFace *face,
                                          int facenum, void *param)
{
    SCE_SMeshFaceInfo *fi;
    fi = SCE_List_GetData (SCE_List_GetIterator (mesh->sortedfaces, facenum));
    fi->index[0] = face->index[0];
    fi->index[1] = face->index[1];
    fi->index[2] = face->index[2];
    fi->index[3] = face->index[3];
    return SCE_TRUE;
}

/* construit la liste des faces triees et la met a jour.
   procede uniquement a sa mise a jour si elle est deja cree */
static int SCE_Mesh_MakeSortedList (SCE_SMesh *mesh)
{
    int diff;
    if (!mesh->sortedfaces)
    {
        mesh->sortedfaces = SCE_List_Create (SCE_Mesh_DeleteFaceInfo);
        if (!mesh->sortedfaces)
        {
            Logger_LogSrc ();
            return SCE_ERROR;
        }
    }
    /* calcul de la difference entre le nombre reel de faces
       et la taille de la sorted list */
    diff = SCE_Mesh_GetNumFaces (mesh) - SCE_List_GetSize (mesh->sortedfaces);
    /* s'il y a plus de faces que ne peut en contenir la sorted list, on l'agrandie */
    if (diff > 0)
    {
        unsigned int i;
        for (i=0; i<diff; i++)
        {
            SCE_SMeshFaceInfo *fi = SCE_malloc (sizeof *fi);
            if (!fi || SCE_List_AddBack (mesh->sortedfaces, fi) < 0)
            {
                SCE_free (fi);
                Logger_LogSrc ();
                return SCE_ERROR;
            }
        }
    }
    /* copie brut des donnees */
    if (SCE_Mesh_ForEachFace (mesh, SCE_Mesh_CopyFaceToSortedList, NULL) < 0)
    {
        Logger_LogSrc ();
        return SCE_ERROR;
    }
    return SCE_OK;
}

/* callbacks */
static int SCE_Mesh_OrderNearToFar (SCE_SMeshFaceInfo *first,
                                    SCE_SMeshFaceInfo *next)
{
    return (first->dist > next->dist);
}
static int SCE_Mesh_OrderFarToNear (SCE_SMeshFaceInfo *first,
                                    SCE_SMeshFaceInfo *next)
{
    return (first->dist < next->dist);
}

/* ajoute le 10/07/2008 */
static void SCE_Mesh_SortSortedList (SCE_SMesh *mesh, int order)
{
    SCE_SListIterator *it = NULL, *jt = NULL;
    SCE_SMeshFaceInfo *info = NULL, *tmp = NULL;
    int (*needChange)(SCE_SMeshFaceInfo*, SCE_SMeshFaceInfo*);

    if (order == SCE_MESH_SORT_NEAR_TO_FAR)
        needChange = SCE_Mesh_OrderNearToFar;
    else
        needChange = SCE_Mesh_OrderFarToNear;

    /* pour chaque face */
    SCE_List_ForEach (it, mesh->sortedfaces)
    {
        info = SCE_List_GetData (it);
        jt = it;
        /* on verifie dans chacune des faces suivantes */
        SCE_List_ForEachNext (jt)
        {
            tmp = SCE_List_GetData (jt);
            /* s'il faut les inverser */
            if (needChange (info, tmp))
            {
                /* si oui, alors on echange les donnees des iterateurs */
                SCE_List_SetData (jt, info);
                SCE_List_SetData (it, tmp);
                /* et on remplace la face que l'on compare par la nouvelle */
                info = tmp;
            }
        }
    }
}

/* ajoute le 27/03/2008 */
int SCE_Mesh_SortFaces (SCE_SMesh *mesh, int order, SCE_TVector3 campos)
{
    unsigned int i, j = 0;
    SCE_SListIterator *it = NULL;
    SCE_SMeshFaceInfo *info = NULL;
    SCE_TVector3 center, vertex[4];
    int vpf;
    SCEvertices *pos = mesh->pos;
    SCEindices *indices = mesh->ib.data;

    /**
     * \todo: IMPORTANT: le mesh doit imperativement posseder des indices
     */

    /* TODO: condition merdique */
    if (!mesh->sortedfaces)
        if (SCE_Mesh_MakeSortedList (mesh) < 0)
        {
            Logger_LogSrc ();
            return SCE_ERROR;
        }

    vpf = SCE_Mesh_GetNumVerticesPerFace (mesh);

    /* calcul de la distance de chaque triangle par rapport a la camera */
    SCE_List_ForEach (it, mesh->sortedfaces)
    {
        info = SCE_List_GetData (it);
        for (i=0; i<vpf; i++)
            SCE_Vector3_Copy (vertex[i], &pos[info->index[i]]);
        SCE_Mesh_GetFaceCenter (vertex, vpf, center);
        info->dist = SCE_Vector3_Distance (center, campos);
    }

    /* tri des faces */
    SCE_Mesh_SortSortedList (mesh, order);

    i = 0;
    /* mise a jour des donnees dans mesh->ib.data */
    SCE_List_ForEach (it, mesh->sortedfaces)
    {
        info = SCE_List_GetData (it);
        for (j=0; j<vpf; j++)
            indices[i+j] = info->index[j]/3;
        i += vpf;
        /* TODO: d'apres l'effet de MakeSortedList, il peut y avoir un
           debordement de 'data' ici (sortedfaces n'est jamais raccourcie) */
    }
    /* mise a jour du IBO */
    /* TODO: proceder a un log update... hqn ah non pas possible :D */
    SCE_CLockIndexBuffer (mesh->ib.buffer, GL_WRITE_ONLY);
    SCE_CUpdateIndexBuffer (mesh->ib.buffer);
    SCE_CUnlockIndexBuffer (mesh->ib.buffer);

    return SCE_OK; /* we won */
}


/* revise le 23/02/2008 */
/* calcul les tangentes, normales et binormales d'un triangle.
   index represente un tableau de 3 indices qui seront utilises
   pour dereferencer les tableaux (tous)
   tangents, binormals et/ou normals peut etre NULL,
   dans quel cas il ne sera pas calcule */
void SCE_Mesh_ComputeTriangleTBN (
  SCEvertices *vertex,
  SCEvertices *texcoord,
  SCEindices *index,

  SCEvertices *tangents,
  SCEvertices *binormals,
  SCEvertices *normals)
{
    /* vecteurs des 'cotes' de la face */
    SCE_TVector3 side0, side1;
    /* vecteurs temporaires */
    SCE_TVector3 tmpTangente, tmpNormal, tmpBinormal;
    SCEvertices deltaT0=0., deltaT1=0., deltaB0=0., deltaB1=0., scale=0.;
    SCEindices default_indices[3] = {0, 1, 2};

    SCE_btstart ();
    if (!index)
        index = default_indices;

    /* on calcule 2 vecteurs formant les bords du triangle
       le vecteur side0 est celui allant du vertex 0 au vertex 1 */
    SCE_Vector3_Operator2v (side0, =, &vertex[index[1]*3],
                                   -, &vertex[index[0]*3]);
    /* le vecteur side1 est celui allant du vertex 0 au vertex 2 */
    SCE_Vector3_Operator2v (side1, =, &vertex[index[2]*3],
                                   -, &vertex[index[0]*3]);

    if (texcoord)
    {
        /* calculs douteux mais qui parait-il marchent...
           (et c'est verifie, ils marchent) */
        /* ici, +1 pour prendre la composante v (uv) */
        deltaT0 = texcoord[index[1]*2+1] - texcoord[index[0]*2+1];
        deltaT1 = texcoord[index[2]*2+1] - texcoord[index[0]*2+1];
        /* composante u */
        deltaB0 = texcoord[index[1]*2] - texcoord[index[0]*2];
        deltaB1 = texcoord[index[2]*2] - texcoord[index[0]*2];
        scale = 1.0f / ((deltaB0 * deltaT1) - (deltaB1 * deltaT0));
    }

    if (normals)
    {
        SCE_Vector3_Cross (tmpNormal, side0, side1);
        SCE_Vector3_Normalize (tmpNormal);
        /* affectation */
        SCE_Vector3_Operator1v (&normals[index[0]*3], +=, tmpNormal);
        SCE_Vector3_Operator1v (&normals[index[1]*3], +=, tmpNormal);
        SCE_Vector3_Operator1v (&normals[index[2]*3], +=, tmpNormal);
    }

    if (texcoord && tangents)
    {
        /* on calcule la tangente temporaire */
        SCE_Vector3_Operator1v (tmpTangente, = deltaT1*, side0);
        SCE_Vector3_Operator1v (tmpTangente, -= deltaT0*, side1);
        SCE_Vector3_Operator1 (tmpTangente, *=, scale);
        SCE_Vector3_Normalize (tmpTangente);
        /* affectation (addition) */
        SCE_Vector3_Operator1v (&tangents[index[0]*3], +=, tmpTangente);
        SCE_Vector3_Operator1v (&tangents[index[1]*3], +=, tmpTangente);
        SCE_Vector3_Operator1v (&tangents[index[2]*3], +=, tmpTangente);
    }
    if (texcoord && binormals)
    {
        /* on calcule la binormale temporaire */
        SCE_Vector3_Operator1v (tmpBinormal, = -deltaB1*, side0);
        SCE_Vector3_Operator1v (tmpBinormal, += deltaB0*, side1);
        SCE_Vector3_Operator1 (tmpBinormal, *=, scale);
        SCE_Vector3_Normalize (tmpBinormal);
        /* affectation (addition) */
        SCE_Vector3_Operator1v (&binormals[index[0]*3], +=, tmpBinormal);
        SCE_Vector3_Operator1v (&binormals[index[1]*3], +=, tmpBinormal);
        SCE_Vector3_Operator1v (&binormals[index[2]*3], +=, tmpBinormal);
    }
    SCE_btend ();
}


/* revise le 16/10/2007 */
int SCE_Mesh_ComputeTBN (
  SCEenum polygon_type,   /* type des polygones */
  SCEvertices *vertex,    /* positions des sommets */
  SCEvertices *texcoord,  /* coordonnees de texture */
  SCEindices *indices,    /* indices (peut valoir NULL) */
  unsigned int icount,    /* nombre d'indices */
  unsigned int vcount,    /* nombre de sommets */

  SCEvertices *tangents,
  SCEvertices *binormals,
  SCEvertices *normals)
{
    unsigned int i;
    SCEindices *index = NULL; /* indices finaux (index si != NULL sinon i) */
    SCEindices t_indices[3];  /* indices temporaires pour ComputeTriangleTBN */
    unsigned int count;       /* count reel */

    SCE_btstart ();
    /* on suppose que les tableaux tangents, binormals et normals
       sont initialises a des valeurs (vecteurs) nulles (0.0;0.0;0.0) */

    /* creation d'un tableau d'indices temporaire
       si aucun tableau d'indices n'a ete donne */
    if (!indices)
    {
        /* aucun index fourni, creation */
        count = vcount;
        index = SCE_malloc (sizeof *index * count);
        if (!index)
        {
            Logger_LogSrc ();
            SCE_btend ();
            return SCE_ERROR;
        }
        /* tableau croissant */
        for (i=0; i<count; i++)
            index[i] = i;
    }
    else
    {
        /* des indices ont ete donnes, on les utilise */
        index = indices;
        count = icount;
    }

    /* suivant le type de polygone, on construit un triangle
       en suivant des regles bien precises, puis en donne
       notre triangle a manger a notre fonction cherie ComputeTriangleTBN */
    switch (polygon_type)
    {
    /* NOTE: penser a modifier SCECTypes.h en consequence */
    case SCE_TRIANGLES:
        /* la, c'est pepere */
        /* i+=3: pour chaque triangle */
        for (i=0; i<count; i+=3)
            SCE_Mesh_ComputeTriangleTBN (vertex, texcoord, &index[i],
                                         tangents, binormals, normals);
        break;

    case SCE_TRIANGLE_STRIP:
        /* la, ca va encore... */
        for (i=0; i<count-2; i++)
            SCE_Mesh_ComputeTriangleTBN (vertex, texcoord, &index[i],
                                         tangents, binormals, normals);
        break;

    case SCE_TRIANGLE_FAN:
        /* la, c'est deja moins rigolo... */
        /* tous les triangles ont un sommet en commun (le premier) donc : */
        t_indices[0] = index[0];

        /* en commencant par le premier sommet apres le sommet commun,
           jusqu'au dernier sommet -1 */
        for (i=1; i<count-1; i++)
        {
            t_indices[1] = index[i];
            t_indices[2] = index[i+1];
            SCE_Mesh_ComputeTriangleTBN (vertex, texcoord, t_indices,
                                         tangents, binormals, normals);
        }
        break;

    case SCE_QUADS:
        /* O joie, des quads... */
        /* i+=4: pour chaque quads */
        for (i=0; i<count; i+=4)
        {
            /* 1 quads = 2 triangles, donc : */
            t_indices[0] = index[i];
            t_indices[1] = index[i+1];
            t_indices[2] = index[i+2];
            SCE_Mesh_ComputeTriangleTBN (vertex, texcoord, t_indices,
                                         tangents, binormals, normals);

            t_indices[0] = index[i+2];
            t_indices[1] = index[i+3];
            t_indices[2] = index[i];
            SCE_Mesh_ComputeTriangleTBN (vertex, texcoord, t_indices,
                                         tangents, binormals, normals);
        }
        break;

    default:
        /* type de primitive non supporte, dommage pour l'utilisateur :D */
        if (!indices)
            SCE_free (index);
        Logger_Log (SCE_INVALID_OPERATION);
        Logger_LogMsg ("primitive type unsupported, you must choose one"
                       " of the following types: SCE_TRIANGLES, "
                       "SCE_TRIANGLE_STRIP, SCE_TRIANGLE_FAN or SCE_QUADS");
        SCE_btend ();
        return SCE_ERROR;
    }

    /* on libere ce qu'on a (eventuellement) alloue */
    if (!indices)
        SCE_free (index);
    index = NULL;

    /* renormalisation des vecteurs */
    if (normals)
        for (i=0; i<vcount; i++)
            SCE_Vector3_Normalize (&normals[i*3]);

    if (binormals)
        for (i=0; i<vcount; i++)
            SCE_Vector3_Normalize (&binormals[i*3]);

    if (tangents)
        for (i=0; i<vcount; i++)
            SCE_Vector3_Normalize (&tangents[i*3]);

    SCE_btend ();
    return SCE_OK;
}

/* ajoute le 22/02/2008 */
/* renvoie les donnees demandees si elles sont conformes aux demandes
   des fonctions de gestion des donnees */
static SCE_SMeshVertexData* SCE_Mesh_GetVertices (SCE_SMesh *mesh, SCEenum type,
                                                  const char *name)
{
    SCE_SMeshVertexData *v = NULL;
    SCE_btstart ();
    /* on recupere les donnees */
    v = SCE_Mesh_LocateData (mesh, type, NULL);
    if (!v)
    {
        Logger_Log (SCE_INVALID_OPERATION);
        Logger_LogMsg ("any buffer contains %s in this mesh", name);
        SCE_btend ();
        return NULL;
    }
    /* on verifie si les donnees sont du bon type */
    if (v->dec->type != SCE_VERTICES_TYPE)
    {
        Logger_Log (SCE_INVALID_OPERATION);
        Logger_LogMsg ("data type of the %s is not SCE_VERTICES_TYPE", name);
        SCE_btend ();
        return NULL;
    }
    SCE_btend ();
    return v;
}
/* idem que GetVertices, mais pour les indices */
static SCEindices* SCE_Mesh_GetIndices (SCE_SMesh *mesh, int *del)
{
    SCEindices *indices = NULL;
    if (mesh->use_indices)
    {
        if (mesh->ib.type != SCE_INDICES_TYPE)
        {
            indices = SCE_Mem_ConvertDup (SCE_INDICES_TYPE, mesh->ib.type,
                                          mesh->ib.data, mesh->icount);
            *del = SCE_TRUE;
        }
        else
        {
            indices = mesh->ib.data;
            *del = SCE_FALSE;
        }
    }
    return indices;
}

/* revise le 22/02/2008 */
int SCE_Mesh_GenerateTBN (
  SCE_SMesh *mesh,          /* mesh */
  SCEvertices *tangents,    /* tableau des tangentes */
  SCEvertices *binormals,   /* tableau des binormales */
  SCEvertices *normals,     /* tableau des normales */
  SCEuint texunit)          /* unite de texture comportant les
                               coordonnees de la normal-map */
{
    SCE_SMeshVertexData *vertex = NULL, *texcoord = NULL, tmp;
    SCEindices *indices = NULL;
    int del = 0; /* kick the warning "used uninitialized" */

    SCE_btstart ();
    if (texunit >= SCE_CGetMaxTextureUnits ())
    {
        /* TODO: message d'erreur frequent, prevoir qq'chose... */
        Logger_Log (SCE_INVALID_ARG);
        Logger_LogMsg ("the texture unit number %d are not available on your "
                       "hardware", texunit);
        SCE_btend ();
        return SCE_ERROR;
    }

    /* on recupere les donnees */
    vertex = SCE_Mesh_GetVertices (mesh, SCE_POSITION, "vertices positions");
    if (tangents || binormals)
        texcoord = SCE_Mesh_GetVertices (mesh, SCE_TEXCOORD0 + texunit,
                                         "texture coordinates");
    else
    {
        SCE_Mesh_InitVertexData (&tmp);
        texcoord = &tmp;
    }

    if (!vertex || !texcoord)
    {
        Logger_LogSrc ();
        SCE_btend ();
        return SCE_ERROR;
    }
    /* on recupere les indices */
    indices = SCE_Mesh_GetIndices (mesh, &del);

    SCE_Mesh_ComputeTBN (mesh->polygon_type, vertex->data, texcoord->data,
                         indices, mesh->icount, mesh->vcount,
                         tangents, binormals, normals);

    if (del)
        SCE_free (indices);

    SCE_btend ();
    return SCE_OK;
}


/* ajoute le 25/06/2007 */
/* revise le 22/02/2008 */
int SCE_Mesh_AddGenVertices (
  SCE_SMesh *mesh,
  SCEuint buf,              /* id du tampon de destination des donnees */
  SCEenum gen_type,         /* types des donnees que l'on souhaite generer */
  SCEuint attrib_type,      /* type des donnees de sortie */
  SCEuint texunit)          /* unite de texture des coordonnees de texture */
{
    SCEvertices *t = NULL, *b = NULL, *n = NULL;
    unsigned int i;
    size_t nvars;

    SCE_btstart ();
    #define SCE_MESH_DEL()\
    {\
        SCE_free (n);\
        SCE_free (b);\
        SCE_free (t);\
        Logger_LogSrc ();\
        SCE_btend ();\
        return SCE_ERROR;\
    }
    nvars = mesh->vcount * 3;
    /* analyse des types a generer */
    if (gen_type & SCE_GEN_TANGENTS)
    {
        t = SCE_malloc (nvars * sizeof *t);
        if (!t)
            SCE_MESH_DEL ()
        for (i=0; i<nvars; i++)
            t[i] = 0.0f;
    }
    if (gen_type & SCE_GEN_BINORMALS)
    {
        b = SCE_malloc (nvars * sizeof *b);
        if (!b)
            SCE_MESH_DEL ()
        for (i=0; i<nvars; i++)
            b[i] = 0.0f;
    }
    if (gen_type & SCE_GEN_NORMALS)
    {
        n = SCE_malloc (nvars * sizeof *n);
        if (!n)
            SCE_MESH_DEL ()
        for (i=0; i<nvars; i++)
            n[i] = 0.0f;
    }
    /***/

    /* generation */
    if (SCE_Mesh_GenerateTBN (mesh, t, b, n, texunit) < 0)
        SCE_MESH_DEL ()

    /** TODO: attrib_type peut devenir trop grand */
    /* ajout */
    if (gen_type & SCE_GEN_TANGENTS)
    {
        if (SCE_Mesh_AddVertices (mesh, buf, attrib_type, SCE_VERTICES_TYPE,
                                  3, mesh->vcount, t, SCE_TRUE) < 0)
            SCE_MESH_DEL ()
        attrib_type++;
    }
    if (gen_type & SCE_GEN_BINORMALS)
    {
        if (SCE_Mesh_AddVertices (mesh, buf, attrib_type, SCE_VERTICES_TYPE,
                                  3, mesh->vcount, b, SCE_TRUE) < 0)
            SCE_MESH_DEL ()
        attrib_type++;
    }
    if (gen_type & SCE_GEN_NORMALS)
    {
        if (SCE_Mesh_AddVertices (mesh, buf, attrib_type, SCE_VERTICES_TYPE,
                                  3, mesh->vcount, n, SCE_TRUE) < 0)
            SCE_MESH_DEL ()
    }
    /***/

    #undef SCE_MESH_DEL

    SCE_btend ();
    return SCE_OK;
}


/* ajoute le 22/02/2008 */
void SCE_Mesh_ComputeBoundingBox (SCEvertices *v, unsigned int vcount,
                                  SCE_TVector3 o, float *w, float *h, float *d)
{
    SCE_TVector3 max = {0., 0., 0.}, min = {0., 0., 0.};
    unsigned int i, p, count;

    /* TODO: utiliser un "Rectangle3D" ? */
    SCE_btstart ();
    count = vcount * 3;
    for (i=0; i<count; i++)
    {
        p = i % 3;
        max[p] = (v[i] > max[p] ? v[i] : max[p]);
        min[p] = (v[i] < min[p] ? v[i] : min[p]);
    }
    SCE_Vector3_Operator1v (o, =, min);
    *w = max[0] - min[0];
    *h = max[1] - min[1];
    *d = max[2] - min[2];
    SCE_btend ();
}
/* ajoute le 23/02/2008 */
void SCE_Mesh_ComputeBoundingBoxv (SCEvertices *v, unsigned int vcount,
                                   SCE_TVector3 o, SCE_TVector3 vec)
{
    SCE_Mesh_ComputeBoundingBox (v, vcount, o, &vec[0], &vec[1], &vec[2]);
}
/* ajoute le 23/02/2008 */
/* revise le 29/02/2008 */
int SCE_Mesh_GenerateBoundingBox (SCE_SMesh *mesh, SCE_SBoundingBox *box)
{
    SCE_SMeshVertexData *v = NULL;
    SCE_TVector3 o, d;

    SCE_btstart ();
    v = SCE_Mesh_GetVertices (mesh, SCE_POSITION, "vertices positions");
    if (!v)
    {
        Logger_LogSrc ();
        SCE_btend ();
        return SCE_ERROR;
    }
    SCE_Mesh_ComputeBoundingBoxv (v->data, mesh->vcount, o, d);
    SCE_BoundingBox_Setv (box, o, d);
    SCE_btend ();
    return SCE_OK;
}

/* ajoute le 22/02/2008 */
void SCE_Mesh_ComputeBoundingSphere (SCEvertices *v, unsigned int vcount,
                                     SCE_TVector3 center, float *r)
{
    SCE_TVector3 tmp;
    float d;
    unsigned int i;

    SCE_btstart ();
    *r = 0.0f;
    /* calcul du centre */
    SCE_Mesh_ComputeBoundingBoxv (v, vcount, center, tmp);
    /* tmp n'est pas un vecteur complet, il faut l'additionner a 'center'
       pour obtenir sa veritable position */
    SCE_Vector3_Operator2v (center, +=, center, +, tmp);
    SCE_Vector3_Operator1 (center, /=, 2.);

    /* determination du point le plus eloigne du centre (donnera le rayon) */
    for (i=0; i<vcount; i++)
    {
        d = SCE_Vector3_Distance (center, &v[i*3]);
        *r = (*r < d ? d : *r);
    }
    SCE_btend ();
}
/* ajoute le 23/02/2008 */
int SCE_Mesh_GenerateBoundingSphere (SCE_SMesh *mesh, SCE_TVector3 c, float *r)
{
    SCE_SMeshVertexData *v = NULL;

    SCE_btstart ();
    v = SCE_Mesh_GetVertices (mesh, SCE_POSITION, "vertices positions");
    if (!v)
    {
        Logger_LogSrc ();
        SCE_btend ();
        return SCE_ERROR;
    }
    SCE_Mesh_ComputeBoundingSphere (v->data, mesh->vcount, c, r);
    SCE_btend ();
    return SCE_OK;
}


/* ajoute le 23/02/2008 */
SCE_SMesh* SCE_Mesh_CreateIndexedCube(SCE_TVector3 o, float w, float h, float d)
{
    SCEvertices v[] =
    {
        o[0],     o[1],     o[2],
        o[0] + w, o[1],     o[2],
        o[0] + w, o[1] + h, o[2],
        o[0] + w, o[1] + h, o[2] + d,
        
        o[0] + w, o[1],     o[2] + d,
        o[0],     o[1] + h, o[2] + d,
        o[0],     o[1] + h, o[2],
        o[0],     o[1],     o[2] + d
    };
    SCEindices i[] =
    {
        0, 1, 2, 6,
        0, 1, 4, 7,
        0, 6, 5, 7,
        1, 2, 3, 4,
        2, 3, 5, 6,
        3, 4, 7, 5
    };
    SCE_SMesh *mesh = NULL;

#define SCE_ASSERT(c)\
    if (c)\
    {\
        SCE_Mesh_Delete (mesh);\
        Logger_LogSrc ();\
        SCE_btend ();\
        return NULL;\
    }
    SCE_btstart ();
    SCE_ASSERT (!(mesh = SCE_Mesh_Create ()))
    SCE_ASSERT (SCE_Mesh_AddVerticesDup (mesh, 0, SCE_POSITION,
                                         SCE_VERTICES_TYPE, 3, 8, v) < 0)
    SCE_ASSERT (SCE_Mesh_SetIndicesDup (mesh, 0, SCE_INDICES_TYPE, 24, i) < 0)
    SCE_Mesh_ActiveVB (mesh, 0, 1);
    SCE_Mesh_ActiveIB (mesh, 1);
    SCE_ASSERT (SCE_Mesh_Build (mesh) < 0)
    SCE_Mesh_SetRenderMode (mesh, GL_QUADS);
    SCE_btend ();
#undef SCE_ASSERT
    return mesh;
}
/* ajoute le 23/02/2008 */
SCE_SMesh* SCE_Mesh_CreateIndexedCubev (SCE_TVector3 o, SCE_TVector3 d)
{
    return SCE_Mesh_CreateIndexedCube (o, d[0], d[1], d[2]);
}

/* ajoute le 05/04/2008 */
SCE_SMesh* SCE_Mesh_CreateCube (SCE_TVector3 o, float w, float h, float d)
{
    SCEvertices v[] =
    {
        /* face profonde */
        o[0],     o[1],     o[2],
        o[0] + w, o[1],     o[2],
        o[0] + w, o[1] + h, o[2],
        o[0],     o[1] + h, o[2],
        /* bas */
        o[0],     o[1],     o[2],
        o[0] + w, o[1],     o[2],
        o[0] + w, o[1],     o[2] + d,
        o[0],     o[1],     o[2] + d,
        /* gauche */
        o[0],     o[1],     o[2],
        o[0],     o[1] + h, o[2],
        o[0],     o[1] + h, o[2] + d,
        o[0],     o[1],     o[2] + d,
        /* droite */
        o[0] + w, o[1],     o[2],
        o[0] + w, o[1] + h, o[2],
        o[0] + w, o[1] + h, o[2] + d,
        o[0] + w, o[1],     o[2] + d,
        /* haute */
        o[0] + w, o[1] + h, o[2],
        o[0] + w, o[1] + h, o[2] + d,
        o[0],     o[1] + h, o[2] + d,
        o[0],     o[1] + h, o[2],
        /* face proche */
        o[0] + w, o[1] + h, o[2] + d,
        o[0] + w, o[1],     o[2] + d,
        o[0],     o[1],     o[2] + d,
        o[0],     o[1] + h, o[2] + d,
    };
    SCE_SMesh *mesh = NULL;

#define SCE_ASSERT(c)\
    if (c)\
    {\
        SCE_Mesh_Delete (mesh);\
        Logger_LogSrc ();\
        SCE_btend ();\
        return NULL;\
    }
    SCE_btstart ();
    SCE_ASSERT (!(mesh = SCE_Mesh_Create ()))
    SCE_ASSERT (SCE_Mesh_AddVerticesDup (mesh, 0, SCE_POSITION,
                                         SCE_VERTICES_TYPE, 3, 24, v) < 0)
    SCE_Mesh_ActiveVB (mesh, 0, 1);
    SCE_ASSERT (SCE_Mesh_Build (mesh) < 0)
    SCE_Mesh_SetRenderMode (mesh, GL_QUADS);
    SCE_btend ();
#undef SCE_ASSERT
    return mesh;
}
/* ajoute le 05/04/2008 */
SCE_SMesh* SCE_Mesh_CreateCubev (SCE_TVector3 o, SCE_TVector3 d)
{
    return SCE_Mesh_CreateCube (o, d[0], d[1], d[2]);
}


/* ajoute le 23/02/2008 */
/* revise le 29/02/2008 */
#if 0
SCE_SMesh* SCE_Mesh_CreateFromBoundingBox (SCE_SBoundingBox *box)
{
    SCE_SMesh *cube = NULL;
    SCE_TVector3 d;

#define SCE_ASSERT(c)\
    if (c)\
    {\
        Logger_LogSrc ();\
        SCE_btend ();\
        return NULL;\
    }
    SCE_btstart ();
    SCE_BoundingBox_GetDimensionsv (box, d);
    SCE_ASSERT (!(cube=SCE_Mesh_CreateIndexedCubev(
                        SCE_BoundingBox_GetOrigin(box), d)))
    SCE_btend ();
#undef SCE_ASSERT
    return cube;
}
#endif


/* revise le 17/10/2007 */
int SCE_Mesh_Build (SCE_SMesh *mesh)
{
    SCE_SListIterator *i = NULL;

    SCE_btstart ();
    if (mesh->builded)
    {
        SCE_btend ();
        return SCE_OK;
    }

    SCE_List_ForEach (i, mesh->vertices)
    {
        if (SCE_Mesh_BuildVB (SCE_List_GetData (i)) < 0)
        {
            Logger_LogSrc ();
            SCE_btend ();
            return SCE_ERROR;
        }
    }

    if (SCE_Mesh_BuildIB (&mesh->ib) < 0)
    {
        Logger_LogSrc ();
        SCE_btend ();
        return SCE_ERROR;
    }

    mesh->builded = SCE_TRUE;

    SCE_btend ();
    return SCE_OK;
}


/* revise le 17/10/2007 */
void SCE_Mesh_SetRenderMode (SCE_SMesh *mesh, SCEenum mode)
{
    mesh->polygon_type = mode;
}


/* revise le 17/10/2007 */
int SCE_Mesh_Update (SCE_SMesh *mesh, unsigned int bufid)
{
    SCE_SListIterator *i = NULL;
    SCE_SMeshVertexBuffer *b = NULL;

    i = SCE_List_GetIterator (mesh->vertices, bufid);
    if (!i)
    {
        Logger_LogMsg ("any buffer number %u in this mesh", bufid);
        Logger_LogSrc ();
        SCE_btend ();
        return SCE_ERROR;
    }

    b = SCE_List_GetData (i);
    SCE_CLockVertexBuffer (b->buffer, GL_WRITE_ONLY);
    SCE_CUpdateVertexBuffer (b->buffer);
    SCE_CUnlockVertexBuffer (b->buffer);
    SCE_CFinishVertexBufferUpdates ();
    return SCE_OK;
}
/* ajoute le 26/03/2008 */
int SCE_Mesh_LogUpdate (SCE_SMesh *mesh, unsigned int bufid)
{
    SCE_SListIterator *i = SCE_List_GetIterator (mesh->vertices, bufid);
    if (!i)
    {
        Logger_LogMsg ("any buffer number %u in this mesh", bufid);
        Logger_LogSrc ();
        SCE_btend ();
        return SCE_ERROR;
    }

    upit = SCE_List_GetIterator (updates, num_updates);
    /* pas la peine de tester 'upit', on considere que si une mise a jour est
       demandee c'est que l'utilisateur sait ce qu'il fait :D */
    SCE_List_SetData (upit, SCE_List_GetData (i));
    num_updates++;
    return SCE_OK;
}

/* ajoute le 26/03/2008 */
void SCE_Mesh_ApplyUpdates (void)
{
    SCE_SMeshVertexBuffer *buf = NULL;
    SCE_CVertexBuffer *vb = NULL;

    /* pour chaque buffer a mettre a jour */
    SCE_List_ForEachPrev (upit)
    {
        /* on le met a jour */
        buf = SCE_List_GetData (upit);
        vb = buf->buffer;
        SCE_CLockVertexBuffer (vb, GL_WRITE_ONLY);
        SCE_CUpdateVertexBuffer (vb);
    }
    /* pour chaque buffer a mettre a jour */
    SCE_List_ForEachPrev (upit)
    {
        /* on rend invalide le pointeur */
        buf = SCE_List_GetData (upit);
        vb = buf->buffer;
        SCE_CUnlockVertexBuffer (vb);
    }
    SCE_CFinishVertexBufferUpdates ();
    upit = NULL;
    num_updates = 0;
}


/* ajoute le 13/02/2008 */
/* revise le 22/04/2008 */
SCE_SMesh** SCE_Mesh_Load (const char *fname, int *n_meshs)
{
    SCE_SMesh **meshs = NULL;
    int i, unused = 0;
    size_t size;
    char *subname = NULL;

    SCE_btstart ();
    if (!n_meshs)
        n_meshs = &unused;
    size = strlen (fname) + 8;
    subname = SCE_malloc (size);
    if (!subname)
    {
        Logger_LogSrc ();
        SCE_btend ();
        return NULL;
    }

    meshs = SCE_Resource_Load (fname, n_meshs, NULL);
    if (!meshs)
    {
        SCE_free (subname);
        Logger_LogSrc ();
        SCE_btend ();
        return NULL;
    }

    for (i=0; i<*n_meshs; i++)
    {
        memset (subname, '\0', size);
        strcpy (subname, fname);
        strcat (subname, "_lod");
        strcat (subname, SCE_String_Strof (i, 0));
        /* TODO: et la valeur de retour n'est pas analysee ? */
        SCE_Resource_Add (subname, meshs[i]);
    }
    SCE_free (subname);

    SCE_btend ();
    return meshs;
}


/* revise le 17/10/2007 */
void SCE_Mesh_Use (SCE_SMesh *mesh)
{
    SCE_SListIterator *i = NULL;
    SCE_SMeshVertexBuffer *b = NULL;

    m = mesh;
    if (m)
    {
        SCE_List_ForEach (i, m->vertices)
        {
            b = SCE_List_GetData (i);
            if (b->active)
                SCE_CUseVertexBuffer (b->buffer);
        }
        if (m->use_indices)
            SCE_CUseIndexBuffer (m->ib.buffer);
    }
    else
    {
        SCE_CUseIndexBuffer (NULL);
        SCE_CUseVertexBuffer (NULL);
    }
}

/* revise le 09/01/2008 */
void SCE_Mesh_Draw (void)
{
    if (m->use_indices)
        SCE_CDrawIndexedBuffer (m->polygon_type, m->ib.type, 0, m->icount);
    else
        SCE_CDrawBuffer (m->polygon_type, 0, m->vcount);
}

/* revise le 06/01/2007 */
void SCE_Mesh_Render (SCE_SMesh *mesh)
{
    SCE_SListIterator *i = NULL;
    SCE_SMeshVertexBuffer *b = NULL;

    /* TODO: prevoir des pointeurs de fonction (+ rapide) */
    SCE_List_ForEach (i, mesh->vertices)
    {
        b = SCE_List_GetData (i);
        if (b->active)
            SCE_CUseVertexBuffer (b->buffer);
    }

    if (mesh->use_indices)
    {
        SCE_CUseIndexBuffer (mesh->ib.buffer);
        SCE_CDrawIndexedBuffer (mesh->polygon_type, mesh->ib.type, 0,
                                mesh->icount);
    }
    else
        SCE_CDrawBuffer (mesh->polygon_type, 0, mesh->vcount);

    /* on re-utilise le mesh par defaut */
    SCE_Mesh_Use (m);
}

/* ajoute le 16/02/2008 */
void SCE_Mesh_RenderList (SCE_SMesh **meshs, unsigned int start, unsigned int n)
{
    while (start < n)
        SCE_Mesh_Render (meshs[start++]);
}
